/*
 * kgraft_patch_bsc1038564
 *
 * Fix for CVE-2017-8890, bsc#1038564
 *
 *  Upstream commit:
 *  657831ffc38e ("dccp/tcp: do not inherit mc_list from parent")
 *
 *  SLE12-SP2 commit:
 *  10ad0c77997702915386a7432b8c77c11ca2921c
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <net/request_sock.h>
#include <net/inet_connection_sock.h>
#include <net/tcp_states.h>

static void (*kgr_security_inet_csk_clone)(struct sock *newsk, const struct request_sock *req);

static struct {
	char *name;
	char **addr;
} kgr_funcs[] = {
	{ "security_inet_csk_clone", (void *)&kgr_security_inet_csk_clone },
};

/* patched function */
struct sock *kgr_inet_csk_clone_lock(const struct sock *sk,
	const struct request_sock *req, const gfp_t priority)
{
	struct sock *newsk = sk_clone_lock(sk, priority);

	if (newsk) {
		struct inet_connection_sock *newicsk = inet_csk(newsk);

		newsk->sk_state = TCP_SYN_RECV;
		newicsk->icsk_bind_hash = NULL;

		inet_sk(newsk)->inet_dport = inet_rsk(req)->ir_rmt_port;
		inet_sk(newsk)->inet_num = inet_rsk(req)->ir_num;
		inet_sk(newsk)->inet_sport = htons(inet_rsk(req)->ir_num);
		newsk->sk_write_space = sk_stream_write_space;

		/*
		 * Fix CVE-2017-8890
		 *  +2 lines
		 */
		inet_sk(newsk)->mc_list = NULL;

		newsk->sk_mark = inet_rsk(req)->ir_mark;
		atomic64_set(&newsk->sk_cookie,
			     atomic64_read(&inet_rsk(req)->ir_cookie));

		newicsk->icsk_retransmits = 0;
		newicsk->icsk_backoff	  = 0;
		newicsk->icsk_probes_out  = 0;

		/* Deinitialize accept_queue to trap illegal accesses. */
		memset(&newicsk->icsk_accept_queue, 0, sizeof(newicsk->icsk_accept_queue));

		kgr_security_inet_csk_clone(newsk, req);
	}
	return newsk;
}

static int kgr_patch_bsc1038564_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1038564_init(void)
{
	return kgr_patch_bsc1038564_kallsyms();
}
